/* abdpp.pl, Abductive proof procedure */
verapp(mple).

/* version mple
	extensional derivation added
	no abduction of non ground abducibles
*/

/* version mpl for multiple predicate learning
	delta set of the form
	[[abd,cl],...]
	learned clauses of the form
		rule(Head,Body,Name,Epluscovered,EplusAbducedCovered).
	background clauses as before
	abduction for bk clauses:
	[abd,'bk']
*/

/* version n2
	no more extensional derivation
*/

/* version n1
	corrected a bug in the consistency procedure:
		when a literal not(a) is encountered
		if the abd pp for a suceeds, then we discard the branch
		if it fails, then we must discrd the literal.
*/

/* version n
	the negation of abducible predicates does not have to be included
	in the set of abducible predicates nor the ic <-a(X),not(a(X))
*/

/* version m
	background rules specified as normal prolog rules, without
	the predicate rule(..)
*/

/* version l
	procedure for abducible target: also the examples are used in the
	derivation 
*/

/* version k
	corrected a bug of version i:
		the literal to be adbuced is skolemized only if it is positive
		skolem((not(Goal),..) did not work
*/

/* version j
	corrected a bug of the inconsistency procedure:
		even if no cosntraint was violated, the inc suceedeed
		becuase at_least_one_violated with [] succeeded
*/

/* version i
        treatment of non ground abducibles: every free variable is skolemized
		when the consistency is entered
		restriction: only constraint of the type
			<-a(X),... (type2)
		and not of the type 
			<-not(a(X)).... (type1)
		type1 ic could provide instantiations for X, so they should
		be tried first
		
*/

/* version h:
	ensured that non abducibles are resolved before abucibles in the 
		abductive derivation
*/

/* version g
        negation as failure literals for non abducible predicates are
        treated differently: they are not included in the set of abducibles
        the negation of non abd pred has not to be specified among abd nor
        the ic a,not(a) have to be specified for non abd

        but the negation of abd and the relative constraint have to be spec

	added the message when the depth limit is reached in the consistency
*/

/* version f
	inconsistency checking using only rules, not 
		abduced literals
	corrected the loop problem of inconsistency checking
	select_literal_in_1st_ic added in order to simplify the code
*/

/* version e
	literals that must be true according to ics are not abduced:
		inconsistency checking added
	from setof to findall for abducibles non ground
	selection rule in ics: 
		reduce first non abducibles, without choice point
		then abducibles, keeping a choice point, when the ic contains
			only non abd
		assumption: the rules and ic contain first non abd and then
			abd in their bodies
*/

/* Version d: limit on the depth of derivations added */

/* Version c: merge of version b and a

	version for Sicstus #2.1

	1) uniform treatment of user defined abducibles and default abducibles:
	besides the user defined abd and their constraints, the user must  
	include in the abducibles all the default literals not(a) plus all the
	constraints
	<-a,not(a). (version c) Example of input:

	rule(p,[q]).

	abducibles([not(p),q,not(q)]).

	ic([p,not(p)]).
	ic([q,not(q)]).
	
	2) the algorithm is able to find all the alternative explanations
	(version b) */

/* Version b: differences with version a1:
	the selection of the literal to be proved false in the 
	consistency phase is now done by a select predicate which leaves
	other choices open for backtracking
	In this way this algorithm finds all the minimal models !*/

/* Version a1: differences with version a
	no 3): negation of abducibles has not to appear in the list of
	abd.*/
	

/* Version a: differences with version 0
	1) treatment of negation as abduction for predicates
	not abducibile: not(p) and <-p,not(p) have to be included
	2) abducible can appear in the head of clauses
	3) negation of abducibles have to appear in the list of
	abducibles

*/

/* Version 0:
(negation of abducible is dealt implicitly by the
algorithm, therefore the ics expressing <-abd,not(abd) dont have
to be expressed)
falso, devono essere espressi altrimenti quando si fa la query
	not(abd)
viene subito abdotto not(abd) senza passare alla consistency per
abd*/
:-dynamic not/1.

:-use_module(library(lists)).

abducible(A):-
	abducibles(AbdSet),
	memberchk(A,AbdSet),!.


/* Abductive derivation:
solve_abd(Goal,AbdIn,AbdOut,Depth) */

/* normal resolution steps */
solve_abd([],Abd,Abd,_Depth):-!.

solve_abd([[true,_]|R],AbdIn,AbdOut,Depth):-!,
	solve_abd(R,AbdIn,AbdOut,Depth).  

/* depth limit reached */
solve_abd(Goal,_,_,0):-!,
	format("Depth limit reached, goal ~p~N",[Goal]),
        fail.

/* fact in the training set, ver l */
solve_abd([[not(A),_]|B],AbdIn,AbdOut,Depth):-
        current_eminus(Em),
        member(A,Em),
        solve_abd(B,AbdIn,AbdOut,Depth).

solve_abd([[A,_]|B],AbdIn,AbdOut,Depth):-
        current_eplus(Ep),
        member(A,Ep),
        solve_abd(B,AbdIn,AbdOut,Depth).

/* opposite of the fact in the training set: failure*/
solve_abd([[not(A),_]|_B],_AbdIn,_AbdOut,_Depth):-
        current_eplus(Ep),
        member(A,Ep),!,
        fail.

solve_abd([[A,_]|_B],_AbdIn,_AbdOut,_Depth):-
        current_eminus(Em),
        member(A,Em),!,
        fail.

/* resolution step */

/* background clause */
solve_abd([[A,_]|B],AbdIn,AbdOut,Depth):-
	clause(A,BodyAnd),
	and2list(BodyAnd,Body1),
	add_cl_name(Body1,Body,bk),
	Depth1 is Depth - 1,
	insert_nabd_1st(Body,B,B1),
	solve_abd(B1,AbdIn,AbdOut,Depth1).

/* target clause */
solve_abd([[A,_Name]|B],AbdIn,AbdOut,Depth):-
        rule(A,Body1,Name,_Epc,_Epac),
        add_cl_name(Body1,Body,Name),
        Depth1 is Depth - 1,
        insert_nabd_1st(Body,B,B1),
	solve_abd(B1,AbdIn,AbdOut,Depth1).


/* NAF goal: not(A) where A is not abd */
solve_abd([[not(A),Name]|B],AbdIn,AbdOut,Depth):-
        \+(abducible(A)),!,
        solve_con2([[A,Name]],[[not(A),Name]|AbdIn],AbdOut1,Depth),
        delete(AbdOut1,[not(A),Name],AbdOut2),
	solve_abd(B,AbdOut2,AbdOut,Depth).

/* end of normal resolution steps */

/* (1) A  has already been abduced */
solve_abd([[A,Name]|B],AbdIn,AbdOut,Depth):-
	memberchk([A,_],AbdIn),!,
	(memberchk([A,Name],AbdIn)-> 
        % abducible already present for the same clause, don't add it
		solve_abd(B,AbdIn,AbdOut,Depth)
        ;
	% abducible present for a different clause: add it
		solve_abd(B,[[A,Name]|AbdIn],AbdOut,Depth)
        ).


/* (2) A is abducible but neither A nor not(A) have already been abduced */
solve_abd([[not(A),Name]|B],AbdIn,AbdOut,Depth):-
	ground(A),
	abducible(A),!,
	\+(memberchk([A,_],AbdIn)),
	solve_con([not(A),Name],[[not(A),Name]|AbdIn],AbdOut1,Depth),
	solve_abd(B,AbdOut1,AbdOut,Depth).


solve_abd([[A,Name]|B],AbdIn,AbdOut,Depth):-
	ground(A),
	abducible(A),!,
	\+(memberchk([not(A),_],AbdIn)),
% only negation constraints are considered, therefore no consistency has to
% be tried
	solve_abd(B,[[A,Name]|AbdIn],AbdOut,Depth).

/* A not a list */
solve_abd(A,AbdIn,AbdOut,Depth):-
        \+(is_list(A)),!,
        solve_abd([[A,top]],AbdIn,AbdOut,Depth).


/* Consistency derivation:
solve_con([Goal,Name],AbdIn,AbdOut,Depth) */

solve_con([Goal,Name],AbdIn,AbdOut,Depth):-
% only negation constraints are considered
	skolem(Goal),
	neg(Goal,NegGoal),
	solve_con2([[[NegGoal,Name]]],AbdIn,AbdOut,Depth).
				/* V b: modified */


/* v b: predicate that selects one literal from the ic
 	and puts it in the head of ic
the selection is done only if the ic contains only abducibles */

select_literal_in_1st_ic([],[]):-!.

select_literal_in_1st_ic([FirstIc|RestIc],[NewFirstIc|RestIc]):-
	select_literal(FirstIc,NewFirstIc).

select_literal([],[]):-!.

select_literal([[Lit,Name]|RestIcIn],[L|IcOut]):-
	abducible(Lit),!,
	select(L,[[Lit,Name]|RestIcIn],IcOut).

select_literal(Ic,Ic).

/* all the ic are satisfied */
solve_con2([],Abd,Abd,_):-!. 

/* one of the Ic has been violated */
solve_con2([ [] |_],_,_,_):-!,fail.

/* depth limit reached */
solve_con2(Ics,_,_,0):-!,
	format("Depth limit reached, ics ~p~N",[Ics]),
        fail.

/* fact in the training set, ver l   */
/* if L1 is in the training set, delete it */
solve_con2([ [[not(L1),_]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
        current_eminus(Em),
        member(L1,Em),!,% in reality this prevents from finding other solutions
        select_literal(Rest_l,Rest_l1),
        solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).

solve_con2([ [[L1,_]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
        current_eplus(Ep),
        member(L1,Ep),!,
        select_literal(Rest_l,Rest_l1),
        solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).

/* if not(L1) is in the training set, then remove the ic */

solve_con2([ [[L1,_]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
        current_eminus(Em),
        member(L1,Em),!,
        select_literal_in_1st_ic(Rest_ic,Rest_ic1),
        solve_con2(Rest_ic1,AbdIn,AbdOut,Depth).

solve_con2([ [[not(L1),_]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
        current_eplus(Ep),
        member(L1,Ep),!,
        select_literal_in_1st_ic(Rest_ic,Rest_ic1),
        solve_con2(Rest_ic1,AbdIn,AbdOut,Depth).



/* (3) first try to use resolution */
solve_con2([ [[true,_]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-!,
        select_literal(Rest_l,Rest_l1),
        solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).


% background predicate
solve_con2([ [[L1,_]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	findall(NewIc,(
	        clause(L1,BodyAnd),
        	and2list(BodyAnd,Body1),
		add_cl_name(Body1,Body,bk),
		insert_nabd_1st(Body,Rest_l,NewIc)
		),
		NewListIc),
	(NewListIc=[]->
		fail
	;
		!
	),
	appendc(NewListIc,Rest_ic,[ Ic | Rest_ic1]),
	select_literal(Ic,Ic1),
	Depth1 is Depth - 1,
	solve_con2([Ic1 | Rest_ic1],AbdIn,AbdOut,Depth1).   	

% target predicate
solve_con2([ [[L1,NameL1]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
        findall(NewIc,(
                rule(L1,Body1,Name,_Ep,_Epac),
                add_cl_name(Body1,Body,Name),
                insert_nabd_1st(Body,Rest_l,NewIc)
                ),
                NewListIc),
        (NewListIc=[]->
                fail
        ;
                !
        ),
        appendc(NewListIc,Rest_ic,[ Ic | Rest_ic1]),
        select_literal(Ic,Ic1),
        Depth1 is Depth - 1,
	((ground(L1),abducible(L1))-> % L1 was an abducible with a partial definition:
        % add to the set of abducible the negation of all the literals
        % obtained resolving L1 with the partial definition
                findall([L1,NameL1],(
                        rule(L1,_Body,_,_,_)
                        ),
                        ListL1),
                complement_list(ListL1,ListnotL1),
                append(ListnotL1,AbdIn,Abd0),
		remove_duplicates(Abd0,Abd1)
        ;
                Abd1=AbdIn
        ),
        solve_con2([Ic1 | Rest_ic1],Abd1,AbdOut,Depth1).

/* (4) L1 has already been abduced */
solve_con2([ [[L1,_Name]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	memberchk([L1,_],AbdIn),!,
	select_literal(Rest_l,Rest_l1),
% do not add the abducible
	solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).


/* (5) not(L1) has been abduced */
solve_con2([ [[not(L1),Name]| _Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	memberchk([L1,_],AbdIn),!,
	select_literal_in_1st_ic(Rest_ic,Rest_ic1),
	(memberchk([L1,Name],AbdIn)-> 
        % abducible already present for the same clause, don't add it
                solve_con2(Rest_ic1,AbdIn,AbdOut,Depth)
        ;
                solve_con2(Rest_ic1,[[L1,Name]|AbdIn],AbdOut,Depth)
        ).

solve_con2([ [[L1,Name]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	memberchk([not(L1),_],AbdIn),!,
	select_literal_in_1st_ic(Rest_ic,Rest_ic1),
	(memberchk([not(L1),Name],AbdIn)-> 
	% abducible already present for the same clause, don't add it
		solve_con2(Rest_ic1,AbdIn,AbdOut,Depth)
	;
        	solve_con2(Rest_ic1,[[not(L1),Name]|AbdIn],AbdOut,Depth)
	).

/* (6) L1 is abducible but neither L1 nor not(L1) have already been abduced */

/* (6b) start an abductive derivation for not(L1) */

/* ver n1: Negation as Failure literal for abducibles and non-abducibles
NAF of abd is treated in the same way as for non-abd
 */
solve_con2([ [[not(L1),Name]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	ground(L1),% if L1 is non ground, the literal can not be reduced
	% and we have to backtrack on literal selection. Otherwise, L1
	% may succeed with an instantiation even if the constraint is 
	% violated (see example voted)
	solve_abd([[L1,Name]],AbdIn,AbdOut1,Depth),
	% if the derivation succeed, then drop the branch
	% and abduce L1
	select_literal_in_1st_ic(Rest_ic,Rest_ic1),
        solve_con2(Rest_ic1,AbdOut1,AbdOut,Depth).

solve_con2([ [[not(_L1),_Name]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-!,
        % if the derivation fails, then drop the literal
 	select_literal(Rest_l,Rest_l1),
	solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).


/* ver n1:  L1 is abducible, we try to abduce the negation, if it is possible
then we discard the constraint, otherwise we discard the literal */
solve_con2([ [[L1,Name]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	ground(L1),
	abducible(L1),
	solve_abd([[not(L1),Name]],AbdIn,AbdOut1,Depth),
	select_literal_in_1st_ic(Rest_ic,Rest_ic1),
       	solve_con2(Rest_ic1,AbdOut1,AbdOut,Depth).

solve_con2([ [[L1,_Name]|Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
% discard the literal
        abducible(L1),!,
	select_literal(Rest_l,Rest_l1),
	solve_con2([Rest_l1|Rest_ic],AbdIn,AbdOut,Depth).


/* (7) L1 is not abducible and has no successful rule, then drop the branch */
solve_con2([ [[_L1,_]|_Rest_l] |Rest_ic],AbdIn,AbdOut,Depth):-
	select_literal_in_1st_ic(Rest_ic,Rest_ic1),
        solve_con2(Rest_ic1,AbdIn,AbdOut,Depth).


/* insert_nabd_1st: assumption: in Body the non abd are before the abd */

insert_nabd_1st([],Cl,Cl):-!.

insert_nabd_1st([[Nonabd,Name]|RestBody],RestIc,[[Nonabd,Name]|NewRestIc]):-
	\+(abducible(Nonabd)),!,
	insert_nabd_1st(RestBody,RestIc,NewRestIc).

insert_nabd_1st(RestBodyOfAbd,RestIc,NewIc):-
	appendc(RestIc,RestBodyOfAbd,NewIc).
/* when only abducible are lerft, they are added to the bottom of the ic */

appendc([],L,L):-!. % append redefined for optimization: ! added

appendc([H|T],L,[H|T1]):-
	appendc(T,L,T1).

skolem(not(Goal)):-!,
	Goal=..[_|Args],
        skolem1(Args).

skolem(Goal):-
	Goal=..[_|Args],
	skolem1(Args).

skolem1([]):-!.

skolem1([H|T]):-
	(var(H)->
		new_const(H)
	;
		true
	),
	skolem1(T).

new_const(H):-
	retract(new_const_number(N)),
	N1 is N + 1,
	assert(new_const_number(N1)),
	number_chars(N,Nstr),
	append("skolem",Nstr,Nconst),
	name(H,Nconst).

:-dynamic new_const_number/1.
	
new_const_number(0).


and2list(A,[A]):-
	\+(A=(_B,_C)).


and2list((A,R),[A|T]):-
	and2list(R,T).
	
neg(not(Goal),Goal):-!.

neg(Goal,not(Goal)).

pred_of_lit(not(P),P):-!.

pred_of_lit(P,P).

add_cl_name([],[],_Name):-!.

add_cl_name([H|Body1],[[H,Name]|Body],Name):-
	add_cl_name(Body1,Body,Name).

complement_list([],[]).

complement_list([[L1,Name]|ListL1],[[not(L1),Name]|ListnotL1]):-
        complement_list(ListL1,ListnotL1).

